package pds.label;

import java.io.PrintStream;
import java.util.ArrayList;

/**
 * PDSValue is a class that contains a value which can be associated
 * with an element. A value also has a units description and a potential
 * data type.
 *
 * @author      Todd King
 * @author      Planetary Data System
 * @version     1.0, 02/21/03
 * @since       1.0
 */
public class PDSValue {
 	/** Data type is unspecified */	public static final int	TYPE_NONE		= 0;
 	/** A quoted (") string. */		public static final int	TYPE_STRING		= 1;
 	/** A literal (') string. */	public static final int	TYPE_LITERAL	= 2;
 	/** Ordered list of values. */	public static final int	TYPE_ORDERED	= 3;	
 	/** An unordered list of values. */	public static final int	TYPE_UNORDERED	= 4;
 	
 	/** The value assocated with the instance */
 	public String	mValue = "";
 	
 	/** The units of the value */
 	public String	mUnits = "";
 	
 	/** The data type for the value */
 	public int		mType = TYPE_NONE;

 	/** An array containing a PDSValue object for each value following
 	 * 						the equal sign in the element. There may be any number of
 	 *						values.
 	 */
 	public ArrayList	mValueList = new ArrayList();
 	
 	/** Creates and instance of a PDSValue */
 	public PDSValue() {
 	}
 	
    /** 
     * Return the length of the value when formated using PDS standards. 
     * This length includes any delimiters which may surround the value
     * as well as any units (if specified)
     *
     * @since           1.0
     */
	public int length() {
		int len = 0;
		
		switch(mType) {
			case TYPE_STRING:
			case TYPE_LITERAL:
				len += 2;
			break;
		}
		len += mValue.length();
			
		if(mUnits.length() > 0) {
			len += mUnits.length() + 3;
		}
		
		return len;
	}
	
    /** 
     * Output a value formated using PDS standards. 
     * This will output the value delimited with the properly
     * quotation marks for the data type, followed by the 
     * units. If no units are specified no units will be output.
     *
     * @since           1.0
     */
	public void print() {
		print(System.out, 0, 4, 78);
	}
	
    /** 
     * Output a value formated using PDS standards. 
     * This will output the value delimited with the properly
     * quotation marks for the data type, followed by the 
     * units. If no units are specified no units will be output.
     *
     * @param out    	the stream to print the element to.
     * @param offset    the number of spaces the first character of the string
     *                  is to be offset. 
     * @param indentLength the number of spaces to indent each line not on the 
     * 					same line as the keyword.
     * @param maxLength the maximum length for any line.
     *
     * @since           1.0
     */
	public void print(PrintStream out, int offset, int indentLength, int maxLength) {
		String	buffer;
		PDSValue	value;
		int		col;
		
		switch(mType) {
			case TYPE_STRING:
				out.print("\"");
			break;
			case TYPE_LITERAL:
				out.print("'");
			break;
			case TYPE_ORDERED:
				if(mValueList.size() > 1) out.print("{ ");
			break;
			case TYPE_UNORDERED:
				if(mValueList.size() > 1) out.print("( ");
			break;
		}
		
		
		if(mType == TYPE_ORDERED || mType == TYPE_UNORDERED) {
			out.println("Number of values: " + mValueList.size());
			col = offset;
			for(int i = 0; i < mValueList.size(); i++) {
				value = (PDSValue) mValueList.get(i);
				if(i != 0) {
					out.print(", ");
					col += 2;
				}
				if(!value.isQuoted()) {
					if(col + value.length() > maxLength) { out.print("\r\n"); col = offset; printSpaces(out, col); }
					indentLength = 4;
				}
				col += value.length();
				value.print(out, offset, indentLength, maxLength);
			}
		} else {
			if(isQuoted()) buffer = mValue;
			else buffer = wrapPad(mValue, offset, indentLength, maxLength);
			out.print(buffer);
		}
		
		switch(mType) {
			case TYPE_STRING:
				out.print("\"");
			break;
			case TYPE_LITERAL:
				out.print("'");
			break;
			case TYPE_ORDERED:
				if(mValueList.size() > 1) out.print(" }");
			break;
			case TYPE_UNORDERED:
				if(mValueList.size() > 1) out.print(" )");
			break;
		}
			
		if(mUnits.length() > 0) {
			out.print(" <" + mUnits + ">");
		}
	}
	
    /** 
     * Format the value and return a string containing it.
     *
     * @return		a String contining a formatted representation of the value.
     *
     * @since           1.0
     */
	public String formatValue() {
		return formatValue(false);
	}
	
    /** 
     * Format the value and return a string containing it.
     *
     * @param plain		flag indicating if the value is not to be adorned
 	 *                  with appropriate quotation marks.
 	 *
	 * @return		a String contining a formatted representation of the value.
     *
     * @since           1.0
     */
	public String formatValue(boolean plain) {
		String	buffer = "";
		
		buffer = mValue;
		
		if(!plain) {
			switch(mType) {
				case TYPE_STRING:
					buffer = "\"" + buffer + "\"";
				break;
				case TYPE_LITERAL:
					buffer = "'" + buffer + "'";
				break;
			}
		}
			
		if(mUnits.length() > 0) {
			buffer += " <" + mUnits + ">";
		}
		
		return buffer;
	}

    /** 
     * Returns true is the data type of the value is quoted.
     *
     * @return          true if a quoted data type, false otherwise
     *
     * @since           1.0
     */
	public boolean isQuoted() {
		switch(mType) {
			case TYPE_STRING:
			case TYPE_LITERAL:
				return true;
		}
		return false;
	}
		
    /** 
     * Reformats a string so that it does not exceed 
     * a given length. Also pads the beginning of the line
     * with some number of spaces.
     *
     * @param text    	the string of text to wrap and pad.
     * @param offset    the number of spaces the first character of the string
     *                  is to be offset. 
     * @param indentLength the number of spaces to indent each line not on the 
     * 					same line as the keyword.
     * @param maxLength the maximum length for any line.
     *
     * @return          the newly formatted string.
     *
     * @since           1.0
     */
	public String wrapPad(String text, int offset, int indentLength, int maxLength) {
		String	buffer;
		String[]	line;
		String	indent = "";
		String	result = "";
		String	newline = "\r\n";
		int		i, j, n;
		
		for(i = 0; i < indentLength; i++) indent += " ";
		
		buffer = "";
		line = text.split(newline);
		for(i = 0; i < line.length; i++) {
			line[i] = line[i].trim();
			if(line[i].length() > 0) {
				if(buffer.length() > 0) buffer += " ";
				buffer += line[i];
				n = buffer.length();
				while(n+offset+indentLength > maxLength) {
					for(j = maxLength-offset-indentLength; j > 0; j--) {
						if(buffer.charAt(j) == ' ') {	// Break it
							if(offset == 0) result += indent;
							result += buffer.substring(0, j);
							result += newline;
							buffer = buffer.substring(j+1);
							break;
						}
					}
					n = buffer.length();
					offset = 0;
				}
			} else {	// Blank line 
				if(buffer.length() > 0) {
					if(offset == 0) result += indent;
					result += buffer + newline;
				}
				if(line.length > 1) result += newline;
				buffer = "";
				offset = 0;
			}
		}
		if(buffer.length() > 0) {	// Something left to add
			if(offset == 0) result += indent;
			result += buffer;
		}
		
		return result;
	}

    /** 
     * Print a string of spaces to an output stream. 
 	 * 
     * @param out    	the stream to print the element to.
     * @param count     the number of spaces to print.
     *
     * @since           1.0
     */
	public void printSpaces(PrintStream out, int count) {
		for(int i = 0; i < count; i++) out.print(" ");
	}
	
    /** 
     * Removes all values. 
 	 * 
     * @since           1.0
     */
	public void clear() {
		mValueList.clear();
		mValue = "";
		mUnits = "";
	}

    /** 
     * Parses a string as a value according to the PDS 
	 * Object Defnition Language (ODL). In this context a value
	 * can be a string, literal or list of values. Lists may
	 * be ordered (enclosed in parenthesis () )  or unordered (enclosed
	 * in curly braces {} ). A value may extend over more than 
	 * one physical line if it is quoted or part of a value set.
	 * 
     * @param buffer    the string to parse as a value.
     *
     * @return          <code>true</code> if an value was parsed from properly;
     *                  <code>false</code> if the an error was encountered.
     *
     * @since           1.0
     */
	public boolean parse(char[] cbuff, int startAt) 
		throws PDSException
	{
		
		char		c = 0;
		int			quote = ' ';
		int			i;
		boolean		inQuote = false;
		boolean		inComment = false;
		boolean		inUnits = false;
		boolean		inList = false;
		
		PDSValue	value;
		int			maxValBuff = 1024;
		char[]		valBuff = new char[maxValBuff];
		int			valN = 0;

		System.out.println("Parsing: " + new String(cbuff));
		
		clear();
		value = this;
		
		if(strlen(cbuff) <= startAt || startAt >= strlen(cbuff)) {	// No value
			throw(new PDSException("Empty value at line " + mLineCount));
		}
		
		// Optimization for long quoted strings
		if(cbuff[startAt] == '"' && cbuff[startAt + strlen(cbuff, startAt) - 1] == '"') {
			value.mValue = new String(cbuff, startAt+1, strlen(cbuff, startAt)-2);
			value.mType = PDSValue.TYPE_STRING;
			return true;
		}
		
		// Parse all other types
		valN = 0;		
		for(i = startAt; i < cbuff.length; i++) {
			c = cbuff[i];
			if(c == '\0') break;	// End of line
			
			// Quoted values
			if(inQuote && c == quote) { inQuote = false; quote = ' '; continue; }
			if(!inQuote && Character.isWhitespace(c)) continue;	// Ignore white space
			
			if(!inQuote && (c == '"' || c == '\'')) { 
				inQuote = true; 
				if(c == '"')	{ quote = '"'; value.mType = PDSValue.TYPE_STRING; }
				if(c == '\'')	{ quote = '\''; value.mType = PDSValue.TYPE_LITERAL; }
				continue;
			}
			if(inQuote) { if(valN < maxValBuff) {valBuff[valN++] = c;} continue; }
			
			// Delimited elements
			if(inList && (c == '{' || c == '('))  { // Start of new list
			    newValue = new PDSValue();
			    mValueList.add(newValue);
			    newValue.parse(cbuff, i);
				continue;
			}
			if(c == '{') { value.mType = TYPE_ORDERED; inList = true; continue; }
			if(c == '(') { value.mType = TYPE_UNORDERED; inList = true; continue; }
			
			if(c == '}' || c == ')') { // End of list markers
				if(inList) {
					inList = false;
					return;
				} else {
					throw(new PDSException("Syntax error at line " + mLineCount));
				}
			}
					
			if(c == '<') { inUnits = true; continue; }
			if(c == '>') { inUnits = false; continue; }
			
			// Parse value lists
			if(!inUnits && c == ',') { // Create new value
				value.mValue = new String(valBuff, 0, valN);
				valN = 0;
				value = new PDSValue(); 
				mValueList.add(value); 
				continue; 
			}
			
			if(inUnits) value.mUnits += c;
			else { if(valN < maxValBuff) valBuff[valN++] = c; }
		}
		
		// If something in the val buffer - set value	
		if(valN > 0) {
			value.mValue = new String(valBuff, 0, valN);
		}
		
		return true;
	}
	
    /** 
     * Parses a string as a value according to the PDS 
	 * Object Defnition Language (ODL). In this context a value
	 * can be a stirng, literal or list of values. Lists may
	 * be ordered (enclosed in parenthesis () )  or unordered (enclosed
	 * in curly braces {} ). A value may extend over more than 
	 * one physical line if it is quoted or part of a value set.
	 * 
     * @param buffer    the string to parse as a value.
     *
     * @return          <code>true</code> if an value was parsed from properly;
     *                  <code>false</code> if the an error was encountered.
     *
     * @since           1.0
     */
	public boolean parse(String buffer) 
		throws PDSException
	{
		int			len = buffer.length();
		char[]		clist;
		
		if(len == 0) {
			throw(new PDSException("Empty value at line " + mLineCount));
		}
		
		clist = new char[len];
		buffer.getChars(0, len, clist, 0);
		
		return parseValue(clist, 0);
	}
	
    /** 
     * Returns the number of values in the value list. 
     * An element may contain any number of values, including 0.
     * For example, an element that is just comment text will 
     * have a value size of 0. 
	 * 
     * @return          the number of values in the value list.
     * @since           1.0
     */
	public int valueSize() {
		if(mValueList.size == 0) return 1;	// Self
		return mValueList.size();
	}
	
    /** 
     * Returns the value associated with the value item in the 
     * value array that is assocaited with the given index.
     * If the index is out of range a blank value is returned.
	 * 
     * @param index     the index of the value to return.
     * @return          a string containing the value associated with the given index. 
     * @since           1.0
     */
	public String value(int index) {
		PDSValue	value;
		
		if(mValueList.size() ==0) return mValue;
		if(index < 0 || index >= mValue.size()) return "";
				
		value = (PDSValue) mValueList.get(index);
		return value.mValue;
	}
	
    /** 
     * Returns the units associated with the value item in the 
     * value array that is assocaited with the given index.
     * If the index is out of range a blank value is returned.
	 * 
     * @param index     the index of the value to return.
     * @return          a string containing the units associated with the given index. 
     * @since           1.0
     */
	public String units(int index) {
		PDSValue	value;
		
		if(mValueList.size() == 0) return mUnits;
		if(index < 0 || index >= mValue.size()) return "";
		
		value = (PDSValue) mValueList.get(index);
		return value.mUnits;
	}	

    /** 
     * Create a string representation of the value.
 	 * 
     * @return			a String containing the formatted value.
     *
     * @since           1.0
     *
     */
	public String valueString() {
		return valueString(false, false);
	}
	
    /** 
     * Create a string representation of the value.
 	 * 
 	 * @param plain		flag indicating if the value is not to be adorned
 	 *                  with appropriate quotation marks.
 	 *
     * @return			a String containing the formatted value.
     *
     * @since           1.0
     *
     */
	public String valueString(boolean plain) {
		return valueString(plain, true);
	}
	
    /** 
     * Create a string representation of the value.
 	 * 
 	 * @param plain		flag indicating if the value is not to be adorned
 	 *                  with appropriate quotation marks.
 	 * @param odl		flag indicating if the value is to be formated with
 	 *					ODL syntax for lists.
 	 *
     * @return			a String containing the formatted value.
     *
     * @since           1.0
     *
     */
	public String valueString(boolean plain, boolean odl) {
		String	buffer;
		PDSValue	value;
		int		i, n;
		
		// Format value
		buffer = "";
		n = mValue.size();
		if(n != 0) {
			// Format each value
			for(i = 0; i < n; i++) {
				if(i != 0) buffer += ", ";
				value = (PDSValue) mValue.get(i);
				buffer += value.formatValue(plain);
			}
			
			if(n > 1 && odl) {	// If more than one value in list
				switch(mType) {
					case TYPE_ORDERED:
						buffer = "{" + buffer + "}";
					break;
					case TYPE_UNORDERED:
						buffer = "(" + buffer + ")";
					break;
				}
			}
		}
		return buffer;
	}

	public int strlen(char[] cbuff)
	{
		return strlen(cbuff, 0);
	}
	
	public int strlen(char[] cbuff, int startAt)
	{
		int	total = cbuff.length;
		for(int i = startAt; i < total; i++) {if(cbuff[i] == '\0') return i-startAt; }
		return cbuff.length-startAt;
	}
    
    /** 
     * Dump all information about an element.
 	 * 
     * @param out    	the stream to print the element to.
     *
     * @since           1.0
     */
	public void dump(PrintStream out) {
		PDSValue	value;
		
		out.print("Type: ");
		switch(mType) {
		case TYPE_NONE:
			 out.println("NONE");
			 break;
		case TYPE_LITERAL:
			 out.println("LITERAL");
			 break;
		case TYPE_STRING:
			 out.println("STRING");
			 break;
		case TYPE_ORDERED:
			 out.println("ORDERED");
			 break;
		case TYPE_UNORDERED:
			 out.println("UNORDERED");
			 break;
		}
		
		if(mType == TYPE_ORDERED || mType == TYPE_UNORDERED) {
			out.println("Number of values: " + mValueList.size());
			for(int i = 0; i < mValueList.size(); i++) {
				value = (PDSValue) mValueList.get(i);
				value.dump(out);
			}
		} else {
			out.println("Value: " + mValue);
			out.println("Units: " + mUnits);
		}
	}
} 
 