package pds.util;

import java.io.*;

/**
 * PPITable is a class for reading and writing ASCII tables.
 *
 *
 * @author      Todd King
 * @author      Planetary Data System
 * @version     1.0, 05/19/03
 * @since       1.0
 */
public class PPITable {
	/** base file object */		FileInputStream	mFile = null;
	/** File object reader */	BufferedReader mReader = null;
	/** Array of field names */	String[]	mFieldName;
	/** Array of field Values */	String[]	mFieldValue;
	/** Current record count */	int	mRecordAt = 0;
	
    /** 
     * Creates an instance.
	 *
     * @since           1.0
     */
 	public PPITable() {
	}
	
	/**
	 * Open a table containing ASCII comma seperated values (CSV).
	 * Establishes a connection to a file that contains a table.
	 * The table may contain a header line which contains field
	 * names. 
	 *
	 * @param pathName	the path and name of the file containing the table of data.
	 * @param nameHeader	if <code>true</code> the first line of the file
	 *						is read, parsed and used to define the name of each
	 *						field.. If <code>false</code> no header line is read
	 *						and no names are defined for each field.
	 * 
	 * @return		<code>true</code> if the file was opened.
	 *				<code>false</code> otherwise.
	 *
	 * @since		1.0
	 */
	 public boolean open(String pathName, boolean nameHeader) {
	 	String	buffer;
	 	
	 	mRecordAt = 0;
		// Open the table
		try {
			mFile = new FileInputStream(pathName);
		} catch(Exception e) {
			return false;
		}
		
		// First line contains field names
		mReader = new BufferedReader(new InputStreamReader(mFile));
		if(nameHeader) {
			try {
				buffer = mReader.readLine();
			} catch(Exception e) {
				return false;
			}
	
			mFieldName = parseLine(buffer);
			for(int i = 0; i < mFieldName.length; i++) {
				System.out.println("field[" + i + "]: " + mFieldName[i]);
			}
		}
		
		return true;
	 }
	 
	 
	/**
	 * Close a connection.
	 *
	 * @return 		<code>true</code> if the connection to the source
	 *				was closed properly.
	 *				<code>false</code> otherwise.
	 * 
	 * @since		1.0
	 */
	 public boolean close() {
		try {
			mReader.close();
			mFile.close();
		} catch(Exception e) {
			return false;
		}
		return true;
	 }
		
	/**
	 * Read one record from a connection and parse the record into fields.
	 *
	 * @return 		<code>true</code> if read was successful
	 *				<code>false</code> otherwise.
	 * 
	 * @since		1.0
	 */
	 public boolean readRecord() {
	 	String	buffer;
	 	
		try {
			buffer = mReader.readLine();
		} catch(Exception e) {
			return false;
		}
		if(buffer == null) return false;
		
		mRecordAt++;
		mFieldValue = parseLine(buffer);
		return true;
	 }
	 
	/**
	 * Return the count of the number of records read from the table.
	 *
	 * @return		the count of the number of records read from the table.
	 * 
	 * @since		1.0
	 */
	 public int recordAt() {
	 	return mRecordAt;
	 }
	 
	/**
	 * Parse a string, using a comma (",") as the delimiter.
	 * After parsing the string removes any leading and trailing 
	 * spaces around each constitute part.
	 *
	 * @return		an array of String containing the constitute parts.
	 * 
	 * @since		1.0
	 */
	 public String[] parseLine(String buffer) {
	 	String[]	parts;
	 	
	 	parts = buffer.split(",");
		for(int i = 0; i < parts.length; i++) {
			parts[i] = parts[i].trim();
		}
		return parts;
	 }
	 
	/**
	 * Return the value associated with a field given the index of the field.
	 * 
	 * @param index		the index of the field for which to extract the value.
	 *
	 * @return		a String containing the value in the field.					
	 * 
	 * @since		1.0
	 */
	 public String getValue(int index) {
	 	if(index < 0 || index >= mFieldValue.length) return null;
	 	
	 	return mFieldValue[index];
	 }
	 
	/**
	 * Copy a file from one location to another. 
	 * The copy is failure safe in that if the copy fails to complete
	 * the destination is not over written.
	 * 
	 * @param source	the name of the source file.
	 * @param destination	the name of the file to copy the source file to.
	 *
	 * @return		<code>true</code> if the copy is successful.
	 *				<code>false</code> otherwise.
	 * 
	 * @since		1.0
	 */
	 static public String copy(String source, String destination) {
	 	String	path = "";
		try {
			File dir = new File (".");
			path = dir.getCanonicalPath();

			if(destination.charAt(0) != dir.separatorChar) {
				path += dir.separator + destination;
			} else {
				path = destination;
			}
			
			File outFile = new File(path);
			outFile.delete();	// Remove file if it exists
			
			// copy to file
			BufferedReader reader = new BufferedReader(new FileReader(source));
			BufferedWriter writer =	new BufferedWriter(new FileWriter(path));
			
			// Now copy the bytes
			char[]	buffer = new char[1024];
			int		n;
			
			while ((n = reader.read(buffer, 0, 1024)) > 0) {
				writer.write(buffer, 0, n);
			}
			reader.close();
			writer.close();
		} catch (IOException e) {
			String temp;
			temp = "Unable to copy file to: " + path;
			return temp;
		}
		return null;
	}
	
	/**
	 * Remove a file from the file system
	 * 
	 * @param source	the name of the source file.
	 *
	 * @return		<code>true</code> if the copy is successful.
	 *				<code>false</code> otherwise.
	 * 
	 * @since		1.0
	 */
	 static public String delete(String source) {
	 	String	path = "";
		try {
			File outFile = new File(source);
			outFile.delete();	// Remove file if it exists
		} catch (Exception e) {
			String temp;
			temp = "Unable to delete file: " + source;
			return temp;
		}
		return null;
	}
	
	/**
	 * Rename a file.
	 * 
	 * @param source	the name of the source file.
	 *
	 * @return		<code>true</code> if the copy is successful.
	 *				<code>false</code> otherwise.
	 * 
	 * @since		1.0
	 */
	 static public String rename(String source, String destination) {
	 	String	path = "";
		try {
			File dir = new File (".");
			path = dir.getCanonicalPath();

			if(destination.charAt(0) != dir.separatorChar) {
				path += dir.separator + destination;
			} else {
				path = destination;
			}
			
			File outFile = new File(path);
			outFile.delete();	// Remove file if it exists
			
			File inFile = new File(source);
			inFile.renameTo(outFile);
		} catch (Exception e) {
			String temp;
			temp = "Unable to rename file: " + source;
			return temp;
		}
		return null;
	}
	
}